# Copyright (C) 1999, 2000 Jay Beale
# Licensed under the GNU General Public License

package Bastille::API;

## TO DO:
#
#
#   1) Consider making the backup directory structure use the same
#      permissions on the directories that the original directories do, so
#      that undo is as easy as  tar -cpf - $backup_dir | ( cd /;  tar -xpf - )
#   2) Look for more places to insert error handling... 
#
#   3) Document this sucker bigtime...
#
#

## TO DO LATER:
#
#   - Use peterw's suggestion about storing diffs of original and Bastille'd
#     files...



##########################################################################
#
# This module forms the basis for the v1.1 API.
# 
 ##########################################################################

# 
# This module forms the initial basis for the Bastille Engine, implemented
# presently via a Perl API for Perl modules.
#
# This is still under construction -- it is very useable, but not very well
# documented, yet.
#

##########################################################################
#
#                          API Function Listing                               
#
##########################################################################
# The routines which should be called by Bastille modules are listed here,
# though they are better documented throughout this file.
#
# Distro Specific Stuff:
#
#  &GetDistro     - figures out what distro we're running, if it knows it...
#  &ConfigureForDistro - sets global variables based on the distro
#
# TUI Integration Stuff:
#
#  &ReadConfig    - reads the config file generated by the TUI
#
# Logging Specific Stuff:
#
#  &ActionLog    -- logs a message to the bastille-action-log file
#  &ErrorLog     -- logs a message to the bastille-error-log file
#
# Input functions for the old input method...
#
#  &GetYN      * -- accepts a Y/N answer from console, ignoring hashed comments
#  &GetString  * -- accepts a line from console, ignoring commented lines
#
# File open/close/backup functions
#
#  &B_open     * -- opens a file handle and logs the action/error (OLD WAY!)
#  &B_open_plus  -- opens a pair of file handles for the old and new version
#                   of a file; respects logonly flag.  (NEW WAY)
#  &B_close    * -- closes a file handle and logs the action/error (OLD WAY!)
#  &B_close_plus -- closes a pair of file handles opened by B_open_plus,
#                   backing up one file and renaming the new file to the
#                   old one's name, logging actions/errors.  Respects the 
#                   logonly flag -- needs B_backup file.  Finally, sets
#                   new file's mode,uid,gid to old file's...  (NEW WAY)
#  &B_backup_file - backs up a file that is being changed/deleted into the
#                   $GLOBAL_BDIR{"backup"} directory.
#
# Content (Line-based) file modification functions
#
#  &B_blank_file  - blanks a file if it isn't ours, so that we can rewrite it
#  &B_append_line - opens a file, appends a line to it, and closes file
#  &B_insert_line - opens a file, inserts a line after a specified line, closes
#  &B_prepend_line- opens a file, prepends a line to it, and closes file
#  &B_replace_line- replaces one or more lines in a file using regep matching
#  &B_hash_comment_line
#                 - hash - comments one or more lines in a file
#  &B_hash_uncomment_line
#                 - hash - uncomments one or more lines in a file
#  &B_delete_line - deletes one or more lines in a file using regep matching
#
# Content (multiple-line-based) file mod functions (for PeterW's ipchains.pm)
#
#  &B_chunk_append
#
#  &B_chunk_replace
#
#
# Non-content file modification functions
#
#  &B_delete_file - deletes the named file, backing up a copy
#  &B_create_file - creates the named file, if it doesn't exist
#
#  &B_print       - print a line to the named filehandle, logging the action
#  &B_chmod       - chmod a file, logging the action
#  &B_chown       - change only the owner on a file, logging the action
#  &B_chgrp       - change the group owner on a file, logging the action
#  &B_symlink     - create a symlink to a file, creating the undo link
#
# More stuff
#
#  &B_chkconfig_off - removed all S links from rcX.d directory
#  &B_createdir     - make a directory, if it doesn't exist, w/ undo link
#  &B_cp            - copy a file, respecting LOGONLY and PREFIX and undo func.
#  &B_mknod         - wrap mknod with undo and logonly and prefix functionality
#

#
# Note:  GLOBAL_LOGONLY
# 
# All Bastille API functions now check for the existence of a GLOBAL_LOGONLY 
# variable.  When said variable is set, no function actually modifies the 
# system.
#

#
# Note: GLOBAL_PREFIX
#
# All Bastille API functions now prepend, directly or indirectly, the variable
# $GLOBAL_PREFIX to the beginning of all filenames.  By "indirectly," I mean
# that some of the API routines use B_open_plus and B_backup_file, which
# prepend the $GLOBAL_PREFIX for them.
#
# For a module to work safely with the $GLOBAL_PREFIX, without said prefix
# being prepended twice, it must not append the prefix to any filename that
# it passes to the API.
#
# This feature is really quite cool: it allows you to modify a distribution 
# that is mounted from another drive.
#


# Libraries for the Backup_file routine: Cwd and File::Path
use Cwd;    
use Getopt::Std;
use English;
use File::Path;

# Export the API functions listed below for use by the modules.

use Exporter;
@ISA = qw ( Exporter );
@EXPORT = qw( PrepareToRun GetDistro ConfigureForDistro StartLogging 
   ActionLog ErrorLog GetYN GetString B_open B_close B_print B_symlink 
   StopLogging B_Display B_open_plus B_close_plus B_blank_file 
   B_append_line B_insert_line B_prepend_line B_replace_line 
   B_hash_comment_line B_hash_uncomment_line B_delete_line B_create_file 
   B_create_dir B_delete_file B_chmod B_chown B_chgrp B_chkconfig_off 
   B_cp B_place B_mknod B_init_file
   $GLOBAL_PREFIX $GLOBAL_LOGONLY $GLOBAL_DISTRO
   %GLOBAL_BIN %GLOBAL_DIR %GLOBAL_FILE %GLOBAL_MISC
   %GLOBAL_BDIR %GLOBAL_BFILE %GLOBAL_CONFIG
);

@ENV="";
$ENV{CDPATH}=".";
$ENV{BASH_ENV}="";

unless ($GLOBAL_BDIR{"install"}) {
    print "Reading configuration files and such...\n";
    &PrepareToRun;
    print "Done reading config files -- going out to make changes...\n";
}

##############################################################################
#
#                         Bastille directory structure
#
##############################################################################
#
#  /root/Bastille  -- parent directory for Bastille Linux
#
#  /root/Bastille/undo -- directory holding all undo scripts and backups...
#  /root/Bastille/undo/backup -- all the original files that Bastille modifies,
#                                with permissions intact
#
#  /root/Bastille/logs -- directory holding all logs
#
# Each of these directories can be prepended by $GLOBAL_PREFIX, allowing
# you to make no changes to the / filesystem while running Bastille on a 
# target filesystem...
#
##############################################################################



##############################################################################
##############################################################################
##################  Actual functions start here... ###########################
##############################################################################
##############################################################################


###########################################################################
# 
# PrepareToRun sets up Bastille to run.  It checks the ARGV array for
# special options and runs ConfigureForDistro to set necessary file
# locations and other global variables.
#
###########################################################################

sub PrepareToRun {

   ### Make sure we're root!
   if ( $EUID != 0 ) {
      die "${PROGRAM_NAME} must be run as root!\n";
   };


   ### Process command-line arguments...
   getopts ('vr:l:u:', \%options);

   if ($options{l}) {
	   $GLOBAL_BDIR{"log"} = $options{l};
   } else {
      $GLOBAL_BDIR{"log"} = "";
   };

   if ($options{r}) {
	   $GLOBAL_PREFIX = $options{r};
   } else {
      $GLOBAL_PREFIX = "";
   };

   if ($options{u}) {
	   $GLOBAL_BDIR{"undo"} = $options{u};
   } else {
      $GLOBAL_BDIR{"undo"} = "";
   };

   if ($options{v}) {
      $GLOBAL_LOGONLY = 1;
   } else {
      $GLOBAL_LOGONLY = 0;
   };


   ### Directories, as explained in Bastille directory structure...
   # Define Installation Directory
   if (${PROGRAM_NAME} !~ m|^/|) {
      my ${CUR_DIR} = cwd;
      $GLOBAL_BDIR{"install"} = ${CUR_DIR} . '/' . ${PROGRAM_NAME}
   } else {
      $GLOBAL_BDIR{"install"} = ${PROGRAM_NAME}
   };

   # Define Undo Directories
   unless ($GLOBAL_BDIR{"undo"}) {
      $GLOBAL_BDIR{"undo"} = $GLOBAL_BDIR{"install"} . "/undo";
   };
   $GLOBAL_BFILE{"created-files"} = $GLOBAL_BDIR{"undo"} .
      "/undo-created-files";
   $GLOBAL_BFILE{"created-dirs"} = $GLOBAL_BDIR{"undo"} . 
      "/undo-created-dirs";
   $GLOBAL_BFILE{"created-symlinks"} = $GLOBAL_BDIR{"undo"} . 
      "/undo-created-symlinks";
   $GLOBAL_BFILE{"removed-symlinks"} = $GLOBAL_BDIR{"undo"} . 
      "/undo-removed-symlinks";
   $GLOBAL_BDIR{"backup"} = $GLOBAL_BDIR{"undo"} . "/backup";
    
   # Define Log Directories
   unless ($GLOBAL_BDIR{"log"}) {
      $GLOBAL_BDIR{"log"} = $GLOBAL_BDIR{"install"} . "/log";
   };
   $GLOBAL_BFILE{"action-log"} = $GLOBAL_BDIR{"log"} . "/action-log";
   $GLOBAL_BFILE{"error-log"} = $GLOBAL_BDIR{"log"} . "/error-log";

   # Define Config Directories
   unless ($GLOBAL_BDIR{"config"}) {
      $GLOBAL_BDIR{"config"} = $GLOBAL_BDIR{"install"} . "/config";
   };
   $GLOBAL_BFILE{"config"} = $GLOBAL_BDIR{"config"} . "/config";

   # Make any directories that don't exist...
   my @created_dirs;
   foreach my $dir (values %GLOBAL_BDIR) {
      push @created_dirs, (mkpath ($dir,0,0700));
   };

   # Log all of the above...
   &ActionLog ("#\n# Bastile GLOBAL_PREFIX  = $GLOBAL_PREFIX");
   &ActionLog ("# Bastile GLOBAL_LOGONLY = $GLOBAL_LOGONLY\n#");
   &ActionLog ("# Bastille Global Directories:");
   foreach my $dir (keys %GLOBAL_BDIR) {
      &ActionLog ("#   $dir = $GLOBAL_BDIR{$dir}");
   };
   &ActionLog ("# Bastille Global Files:");
   foreach my $file (keys %GLOBAL_BFILE) {
      &ActionLog ("#   $file = $GLOBAL_BDIR{$file}");
   };
   &ActionLog ("# Directories Created By Bastille:");
   foreach my $dir (@created_dirs) {
      &ActionLog ("#   $dir");
   };
    
   # Configure for this distro...
   &ConfigureForDistro || die "Could not configure for this distro!\n";

   # Load in the user's configuration file...
   &ReadConfig;
    
};

###########################################################################
#
# GetDistro checks to see if the target is a known distribution and reports
# said distribution.
#
# This is used through the script, but also by ConfigureForDistro.
# 
#
###########################################################################

sub GetDistro() {

   my ($line,$distro);

   my %{RELEASE_FILES} = (
      'Linux' => [
         {
            'distro'  => 'VA',
            'file'    => '/etc/va-release',
            'regex'   => '^VALinux (\d\w*(\.\d\w*)+)',
            'default' => '6.1'
         },
         {
            'distro'  => 'RH',
            'file'    => '/etc/redhat-release',
            'regex'   => '^Red Hat Linux release (\d\w*(\.\d\w*)+)',
            'default' => '6.2'
         },
         {
            'distro'  => 'MN',
            'file'    => '/etc/mandrake-release',
            'regex'   => '^Mandrake Linux release (\d\w*(\.\d\w*)+)',
            'default' => '7.0'
         },
         {
            'distro'  => 'SW',
            'file'    => '/etc/slackware-version',
            'regex'   => '^(\d\w*(\.\d\w*)+)',
            'default' => '7.0.0'
         }
      ],
      'SunOS' => [
         {
            'distro'  => 'SO',
            'file'    => '/etc/release',
            'regex'   => '^[ 	]*Solaris (\d\w*(\.\d\w*)*)',
            'default' => '7'
         }
      ]
   );
      
   if (! $GLOBAL_DISTRO) {
      chomp ($OS = `uname -s`);
      foreach my ${DISTRO} (@{${RELEASE_FILES{$OS}}}) {
         if (-e $DISTRO->{'file'}) {
            &B_open (RELEASE_FILE, "$GLOBAL_PREFIX/$DISTRO->{'file'}");
            my $line;
            while ($line = <RELEASE_FILE>) {
               if ($line =~ /$DISTRO->{'regex'}/) {
                  $distro = $DISTRO->{'distro'} . $1;
                  last;
               } 
            };
            &B_Close (*RELEASE_FILE);
            if (! $distro) {
               $distro = $DISTRO->{'distro'} . $DISTRO->{'default'};
               &ErrorLog("Couldn't determine release version! Setting to $distro!\n");
            };
            last;
         };
      };
   } else {
      $distro = $GLOBAL_DISTRO;
      &ErrorLog("Couldn't determine release version! Setting to $distro!\n");
   };
   $GLOBAL_DISTRO = $distro;
   $distro;
};


###########################################################################
#
# &ConfigureForDistro configures the API for a given distribution.  This
# includes setting global variables that tell the Bastille API about
# given binaries and directories.
#
# WARNING: If a distro is not covered here, Bastille may not be 100% 
#          compatible with it, though 1.1 is written to be much smarter
#          about unknown distros...
#
###########################################################################

sub ConfigureForDistro {

   my $distro = &GetDistro;
   my $return=1;

   my %DISTRO_PATHS = (
      "default" => {
         "accton"            => "/usr/sbin/accton",
         "bash"              => "/bin/bash",
         "chattr"            => "/usr/bin/chattr",
         "chown"             => "/bin/chown",
         "cp"                => "/bin/cp",
         "crontab"           => "/usr/bin/crontab",
         "diff"              => "/usr/bin/diff",
         "floppy"            => "/mnt/floppy",
         "group"             => "/etc/group",
         "groupadd"          => "/usr/sbin/groupadd",
         "home"              => "/home",
         "httpd.conf"        => "/etc/httpd/conf/httpd.conf",
         "httpd_access.conf" => "/etc/httpd/conf/access.conf",
         "initd"             => "/etc/rc.d/init.d",
         "inittab"           => "/etc/inittab",
         "killall"           => "/usr/bin/killall",
         "lilo"              => "/sbin/lilo",
         "lilo.conf"         => "/etc/lilo.conf",
         "limits.conf"       => "/etc/security/limits.conf",
         "log"               => "/var/log",
         "lprm"              => "/usr/bin/lprm",
         "lpr"               => "/usr/bin/lpr",
         "mknod"             => "/bin/mknod",
         "mount"             => "/bin/mount",
         "mtab"              => "/etc/mtab",
         "mv"                => "/bin/mv",
         "pam_access.conf"   => "/etc/security/access.conf",
         "pamd"              => "/etc/pam.d",
         "rcd"               => "/etc/rc.d",
         "rpm"               => "/bin/rpm",
         "sbin"              => "/sbin",
         "sulogin"           => "/sbin/sulogin",
         "umount"            => "/bin/umount",
         "useradd"           => "/usr/sbin/useradd"
      },
      "MN7.0" => {
      },
      "MN7.1" => {
      },
      "RH5.2" => {
      },
      "RH6.0" => {
         "httpd_access.conf" => "/etc/httpd/conf/httpd.conf"
      },
      "RH6.1" => {
         "httpd_access.conf" => "/etc/httpd/conf/httpd.conf"
      },
      "RH6.2" => {
         "accton" => '/sbin/accton'
      },
      "SO2.6" => {
      },
      "SO7" => {
      },
      "SW7.0.0" => {
         "log"               => "/usr/adm",
         "httpd.conf"        => "/var/lib/apache/conf/httpd.conf",
         "httpd_access.conf" => "/var/lib/apache/conf/access.conf"
      },
      "SW7.1.0" => {
         "log"               => "/usr/adm",
         "httpd.conf"        => "/var/lib/apache/conf/httpd.conf",
         "httpd_access.conf" => "/var/lib/apache/conf/access.conf"
      }
   );
   foreach my $DISTRO_FILE (keys %{$DISTRO_PATHS{'default'}}) {
      my $PATH_TEMP;
      if ($DISTRO_PATHS{$distro}{$DISTRO_FILE}) {
         $PATH_TEMP = $DISTRO_PATHS{$distro}{$DISTRO_FILE};
      } else {
         $PATH_TEMP = $DISTRO_PATHS{'default'}{$DISTRO_FILE};
      };
      if (-d $PATH_TEMP) {
         $GLOBAL_DIR{$DISTRO_FILE} = $PATH_TEMP;
      } elsif (-f $PATH_TEMP && -x $PATH_TEMP) {
         $GLOBAL_BIN{$DISTRO_FILE} = $PATH_TEMP;
      } elsif (-f $PATH_TEMP) {
         $GLOBAL_FILE{$DISTRO_FILE} = $PATH_TEMP;
      } else {
         $return = 0;
      };
   };
   $return;
};


###########################################################################
# &ReadConfig reads in the user's choices from the TUI, stored in the file
# $GLOBAL_BFILE{"config"}.  We were using AppConfig here at first, but it was
# just such a pain in the, ermm, keyboard...
#
###########################################################################

sub ReadConfig {

   open CONFIG, $GLOBAL_BFILE{"config"} 
      or die "Couldn't open config file!\n";
   while (my $line = <CONFIG>) {

      # Skip commented lines...
      unless ($line =~ /^\s*#/) {
         if ($line =~ /^\s*(\w+).(\w+)\s*=\s*\"(.*)\"/ ) {
            $GLOBAL_CONFIG{$1}{$2}=$3;
         };
      }:
   }
   close CONFIG;
};


###########################################################################
# &B_init_file makes sure that a GLOBAL_BFILE file exists before other 
# routines attempt to append to it.
###########################################################################
sub B_init_file {
   for my $FILE (@_) {
      unless ( -e $GLOBAL_BFILE{$FILE} ) {
         open TEMP_FILEHANDLE,">" . $GLOBAL_BFILE{$FILE};
         close TEMP_FILEHANDLE;
         chmod 0700,$GLOBAL_BFILE{$FILE};
         chown 0,0,$GLOBAL_BFILE{$FILE};
      };
   };
};


##############################################################################
# &ActionLog ($Log_items) prints $log_items to the action-log.  If said log
# doesn't exist, it is created.
##############################################################################

sub ActionLog {

   &B_init_file ("action-log");

   open ACTIONLOG,">>" . $GLOBAL_BFILE{"action-log"} 
      or print "Couldn't open action-log!\n";
      foreach my $thing ( @_ ) {
         print ACTIONLOG $thing;
      };
   close ACTIONLOG;
    
};

##############################################################################
# &ErrorLog ($log_items) prints $log_items to the error-log.  If said log
# doesn't exist, it is created.
#
# The idea for this log was Mike Rash's (mbr).
##############################################################################

sub ErrorLog {

   # Write something to our ongoing "errorlog" file

   &B_init_file ("error-log");

   open ERRORLOG,">>" . $GLOBAL_BFILE{"error-log"} 
      or print "Couldn't open error-log!\n";
    
   foreach my $thing ( @_ ) {
      print ERRORLOG $thing;
   };
    
   close ERRORLOG;   
};


###########################################################################
# GetYN and GetString  are the two original v1.0.0-v1.0.3 input functions #
# used to read in (and log) input.                                        #
# Both functions ignore hash-commented lines.                             #
#                                                                         #
# Under the new architecture, these functions should fall out of use      #
# entirely , as we move to a TUI that writes AppConfig files.             #
#                                                                         #
###########################################################################

# &GetYN grabs a Y or an N from stdin.  It 
sub GetYN {
   # enhanced by Don Wilder and Peter W.
   my ( $prefix ) = @_;        # optional prefix
   my ($line, $ok, $warn);
    
   print "Press <Shift><Page Up>/<Page Down> to see previously scrolled text.\n";
   $ok = 0;
   while ( $ok == 0 ) {
      print "$warn$prefix(Y or N): ";
      $line=<STDIN>;
      print INPUTLOG $line;   # log answer
      if ( ($line =~ /^#/) or ($line !~ /^[YN]/i ) ) {
         # If we're in this loop, we've either read a comment, or the
         # input did not begin with Y or N.  Ask again.
         $warn = "Please respond with Y or N. ";
      } else {
         $ok = 1;
      };
   };
   return(uc(substr($line,0,1)));
};

sub GetString {
   my $line;
            
   #  Take the first line that isn't a comment, that is, the first line
   #  not beginning with:      
   #                   whitespace #
   while ( ($line=<STDIN>) =~ /^(\s*)#/) { 
      print INPUTLOG $line;
   };

   print INPUTLOG $line;
   # Now, strip out the comment at the end...

   #
   # Note the strange structure (via the while statement).  This is to
   # prevent weird comments, like:
 
   #                
   #           #  this is a comment # with pound's in the # middle...
   # or 
   #           ##  this is a better example: 
   #      
   #           #  dial  #,#,1,2,3
   # 
    
   while ( $line =~ /^(.*[^\\])#/ ) {
      $pre_pound=$1;
      $pre_pound =~ s/(.*[^\s])\s+/$1/;
      $line = $pre_pound;
   };
   
   # Finally, change all escaped #'s to #'s:   
   $line =~ s/\\#/#/g;

   # Return the found line
   $line;
};


###########################################################################
###########################################################################
#                                                                         #
# The B_<perl_function> file utilities are replacements for their Perl    #
# counterparts.  These replacements log their actions and their errors,   #
# but are very similar to said counterparts.                              #
#                                                                         #
###########################################################################
###########################################################################


###########################################################################
# B_open was the v1.0 open command.  It is still used in places in the
# code, though it should fall out of use quickly in v1.1, as it has been
# replaced by B_open_plus, which implements a new, smarter, backup scheme.
#
# B_open opens the given filehandle, associated with the given filename
# and logs appropriately.
#
# WARNING: B_open does not respect the $GLOBAL_PREFIX -- use this with care!
#
###########################################################################

sub B_open {
   my $return=1;
   my ($handle,$filename)=@_;

   unless ($GLOBAL_LOGONLY) {
      $return = open $handle,$filename;
   };

   ($handle) = "$_[0]" =~ /[^:]+::[^:]+::([^:]+)/;
   &ActionLog("open $handle,\"$filename\";\n");
   unless ($return) {
      &ActionLog("#open $handle , $filename failed...\n");
      &ErrorLog("#open $handle, $filename failed...\n");
   };
   
   $return;
}   

###########################################################################
# B_open_plus is the v1.1 open command.
# 
# &B_open_plus($handle_file,$handle_original,$file) opens the file $file
# for reading and opens the file ${file}.bastille for writing.  It is the
# counterpart to B_close_plus, which will move the original file to
# $GLOBAL_BDIR{"backup"} and will place the new file ${file}.bastille in its
# place.
#
# &B_open_plus makes the appropriate log entries in the action and error
# logs.
###########################################################################

sub B_open_plus {

    my ($handle_file,$handle_original,$file)=@_;
    my $return=1;
    my $return_file=1;
    my $return_old=1;
  
    my $original_file = $file;
    $file .= $GLOBAL_PREFIX;

    # Open the original file and open a copy for writing.
    unless ($GLOBAL_LOGONLY) {
       $return_old = open $handle_original,"$file";
       $return_file = open $handle_file,("> $file.bastille");
    };
    
    # Error handling/logging here...
    #&ActionLog("# Modifying file $original_file via temporary file $original_file.bastille\n");
    unless ($return_file) {
       $return=0;
       &ErrorLog("#open $original_file.bastille failed...\n");
    };
    unless ($return_old) {
       $return=0;
       &ErrorLog("#open $original_file failed.\n");
    };

    $return;
        
};

###########################################################################
# B_close was the v1.0 close command.  It is still used in places in the
# code, though it should fall out of use quickly in v1.1, as it has been
# replaced by B_close_plus, which implements a new, smarter, backup scheme.
#
# B_close closes the given filehandle, associated with the given filename
# and logs appropriately.
###########################################################################


sub B_close {
   my $return=1;

   unless ($GLOBAL_LOGONLY) {
      $return = close $_[0];
   };

   &ActionLog( "close $_[0];\n");
   unless ($return) {
      &ActionLog("#ERROR: close $_[0] failed...\n");
      &ErrorLog( "#ERROR: close $_[0] failed...\n");
   };

   $return;
};


###########################################################################
# B_close_plus is the v1.1 close command.
# 
# &B_close_plus($handle_file,$handle_original,$file) closes the files
# $file and ${file}.bastille, backs up $file to $GLOBAL_BDIR{"backup"} and
# renames ${file}.bastille to $file.  This backup is made using the
# internal API function &B_backup_file.  Further, it sets the new file's
# permissions and uid/gid to the same as the old file.
#
# B_close_plus is the counterpart to B_open_plus, which opened $file and 
# $file.bastille with the file handles $handle_original and $handle_file, 
# respectively.
#
# &B_close_plus makes the appropriate log entries in the action and error
# logs.
###########################################################################

sub B_close_plus {
   my ($handle_file,$handle_original,$file)=@_;
   my ($mode,$uid,$gid);
   my @junk;

   my $original_file;

   my $return=1;
   my $return_file=1;
   my $return_old=1;

   # Append the global prefix, but save the original for B_backup_file b/c
   # it appends the prefix on its own...

   $original_file=$file;
   $file .= $GLOBAL_PREFIX;

   #
   # Close the files and prepare for the rename
   #

   unless ($GLOBAL_LOGONLY) {
      $return_file = close $handle_file;
      $return_old = close $handle_original;
   };

   # Error handling/logging here...
   #&ActionLog("#Closing $original_file and backing up to " . $GLOBAL_BDIR{"backup"});
   #&ActionLog("/$original_file\n");

   unless ($return_file) {
      $return=0;
      &ErrorLog("#close $original_file failed...\n");
   };
   unless ($return_old) {
      $return=0;
      &ErrorLog("#close $original_file.bastille failed.\n");
   };

   #
   # If we've had no errors, backup the old file and put the new one
   # in its place, with the Right permissions.
   #

   unless ( ($return == 0) or $GLOBAL_LOGONLY) {

      # Read the permissions/owners on the old file
      @junk=stat ($file);
      $mode=$junk[2];
      $uid=$junk[4];
      $gid=$junk[5];

      # Set the permissions/owners on the new file
      chmod $mode, "$file.bastille" or &ErrorLog("B_close_plus: Not able to retain permissions on $original_file!!!\n");
      chown $uid, $gid, "$file.bastille" or &ErrorLog("B_close_plus: Not able to retain owners on $original_file!!!\n");

      # Backup the old file and put a new one in place.
   
      &B_backup_file($original_file);
      rename "$file.bastille", $file or &ErrorLog("B_close_plus: not able to move $original_file.bastille to $original_file!!! BAD!!!\n");
    };

    $return;
};

###########################################################################
# &B_backup_file ($file) makes a backup copy of the file $file in 
# $GLOBAL_BDIR{"backup"}.  Note that this routine is intended for internal
# use only -- only Bastille API functions should call B_backup_file.
#
# Note that this routine most _carefully_ handle the GLOBAL_PREFIX variable.
# All routines which call B_backup_file will pass the filename to back up,
# without prepending the prefix.
#
###########################################################################

sub B_backup_file {

   my $file=$_[0];
   my $directory=cwd;

   my $original_file = $file;
   $file .= $GLOBAL_PREFIX;

   my $backup_dir = $GLOBAL_BDIR{"backup"};
   my $backup_file = $backup_dir . $original_file;

   # We don't use this -- this is mostly here for commenting.
   #my $relative_file;

   my $return=1;

   # First, separate the file into the directory and the relative filename

   if ($file =~ /^(.*)\/([^\/]+)$/) {
      #$relative_file=$2;
      $directory = $1;
   };
    

   # Now, if the directory does not exist, create it.
   # Later:
   #   Try to set the same permissions on the patch directory that the
   #   original had...?

   unless ( -d ($backup_dir . $directory) ) {
      mkpath(( $backup_dir . $directory),0,0700);
   };

   # Now we backup the file.  If there is already a backup file there,
   # we will leave it alone, since it exists from a previous run and
   # should be the _original_ (possibly user-modified) distro's version 
   # of the file.

   if ( -e $file ) {
      unless ( -e $backup_file ) {
         `$GLOBAL_BIN{"cp"} -p $file $backup_file`;
      };

      # Later, following PeterW's suggestion, place a diff of the file
      # in a second backup directory

   }   else {
      # The file we were trying to backup doesn't exist.  Note this!
      $return=0;
      &ErrorLog ("# Failed trying to backup file $file -- it doesn't exist!\n");
   };

   # Check to make sure that the file does exist in the backup location.
    
   unless ( -e $backup_file ) {
      $return=0;
      &ErrorLog("# Failed trying to backup $file -- the copy was not created.\n");
   };

   $return;
};


###########################################################################
# &B_blank_file ($filename,$pattern) blanks the file $filename, unless the
# pattern $pattern is present in the file.  This lets us completely redo
# a file, if it isn't the one we put in place on a previous run...
#
# B_blank_file respects $GLOBAL_LOGONLY and uses B_open_plus and B_close_plus
# so that it makes backups and only modifies files when we're not in "-v"
# mode...
#
###########################################################################

sub B_blank_file {
    
   my ($filename,$pattern) = @_;
   my $return;

   # If this variable is true, we won't blank the file...

   my $found_pattern=0;

   if ($return=&B_open_plus (*BLANK_NEW,*BLANK_OLD,$filename) ) {

      my @lines;
   
      while (my $line = <BLANK_OLD>) {
         push @lines,$line;
         if ($line =~ $pattern) {
            $found_pattern=1;
         };
      };

      # Only copy the old file if the new one didn't match.
      if ($found_pattern) {
         while ($line = shift @lines ) {
            &B_print(*BLANK_NEW,$line);
         };
      } else {
         &ActionLog("Blanked file $filename\n");
      };
      &B_close_plus(*BLANK_NEW,*BLANK_OLD,$filename);
   } else {
      &ErrorLog("Couldn't blank file $filename since we couldn't open it or its replacement\n");
   };

};


###########################################################################
# &B_append_line ($filename,$pattern,$line_to_append)  modifies $filename,
# appending $line_to_append unless one or more lines in the file matches
# $pattern.  This is an enhancement to the append_line_if_no_such_line_exists
# idea.
#
# Additionally, if $pattern is set equal to 0, the line is always appended.  
#
# B_append_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
# Here's examples of where you might use this:
#
# You'd like to add a   root   line to /etc/ftpusers if none exists.
# You'd like to add a   Options Indexes  line to Apache's config. file,
# after you delete all Options lines from said config file.
#
###########################################################################

sub B_append_line {
   
   my ($filename,$pattern,$line_to_append) = @_;

   my $found_pattern=0;
   my $return=1;

   if ( &B_open_plus (*APPEND_NEW,*APPEND_OLD,$filename) ) {
      while (my $line=<APPEND_OLD>) {
         &B_print(*APPEND_NEW,$line);
         if ($line =~ $pattern) {
            $found_pattern=1;
         };
      };
      unless ( ($found_pattern) or ($pattern != 0) ) {
         &B_print(*APPEND_NEW,$line_to_append);
         &ActionLog("Appended the following line to $filename:\n");
         &ActionLog("$line_to_append");
      };
      &B_close_plus (*APPEND_NEW,*APPEND_OLD,$filename);
   } else {
      $return=0;
      &ErrorLog("# Couldn't append line to $filename, since open failed.");
   };

   $return;

};


###########################################################################
# &B_insert_line ($filename,$pattern,$line_to_insert,$line_to_follow)  
# modifies $filename, inserting $line_to_insert unless one or more lines
# in the file matches $pattern.  The $line_to_insert will be placed
# immediately after $line_to_follow, if it exists.  If said line does not
# exist, the line will not be inserted and this routine will return 0.
#
# B_insert_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
# Here's examples of where you might use this:
#
# You'd like to insert a line in Apache's configuration file, in a 
# particular section.
#
###########################################################################

sub B_insert_line {
  
   my ($filename,$pattern,$line_to_insert,$line_to_follow) = @_;

   my @lines;
   my $found_pattern=0;
   my $found_line_to_follow=0;

   my $return=1;

   if ( &B_open_plus (*INSERT_NEW,*INSERT_OLD,$filename) ) {

      # Read through the file looking for a match both on the $pattern
      # and the line we are supposed to be inserting after...

      my $ctr=1;
      while (my $line=<INSERT_OLD>) {
         push (@lines,$line);
         if ($line =~ $pattern) {
            $found_pattern=1;
         };
         if ( ($found_line_to_follow < 1) and ($line =~ $line_to_follow)) {
            $found_line_to_follow=$ctr;
         };
         $ctr++;
      };

      # Log an error if we never found the line we were to insert after
      unless ($found_line_to_follow ) {
         $return=0;
         &ErrorLog("Never found the line that we were supposed to insert after in $filename\n");
      };

      # Now print the file back out, inserting our line if we should...

      $ctr=1;
      print "Line to follow is $found_line_to_follow\n";
      while (my $line = shift @lines) {
         &B_print(*INSERT_NEW,$line);
         if ( ($ctr == $found_line_to_follow) and ($found_pattern == 0) ) {
            &B_print(*INSERT_NEW,$line_to_insert);
            &ActionLog("Inserted the following line in $filename:\n");
            &ActionLog("$line_to_insert");
         };
         $ctr++;
      };

      &B_close_plus (*INSERT_NEW,*INSERT_OLD,$filename);
   
   } else {
      $return=0;
      &ErrorLog("# Couldn't prepend line to $filename, since open failed.");
   };

   $return;

};


###########################################################################
# &B_prepend_line ($filename,$pattern,$line_to_prepend)  modifies $filename,
# prepending $line_to_prepend unless one or more lines in the file matches
# $pattern.  This is an enhancement to the prepend_line_if_no_such_line_exists
# idea.  
#
# B_prepend_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
# Here's examples of where you might use this:
#
# You'd like to insert the line "auth   required   pam_deny.so" to the top
# of the PAM stack file /etc/pam.d/rsh to totally deactivate rsh.
#
###########################################################################

sub B_prepend_line {
   
   my ($filename,$pattern,$line_to_prepend) = @_;

   my @lines;
   my $found_pattern=0;
   my $return=1;

   if ( &B_open_plus (*PREPEND_NEW,*PREPEND_OLD,$filename) ) {
      while (my $line=<PREPEND_OLD>) {
         push (@lines,$line);
         if ($line =~ $pattern) {
            $found_pattern=1;
         };
      };
      unless ($found_pattern) {
         &B_print(*PREPEND_NEW,$line_to_prepend);
      };
      while (my $line = shift @lines) {
         &B_print(*PREPEND_NEW,$line);
      };

     &B_close_plus (*PREPEND_NEW,*PREPEND_OLD,$filename);
   
      # Log the action
      &ActionLog("Prepended the following line to $filename:\n");
      &ActionLog("$line_to_prepend");
   } else {
      $return=0;
      &ErrorLog("# Couldn't prepend line to $filename, since open failed.\n");
   };

   $return;

};


###########################################################################
# &B_replace_line ($filename,$pattern,$line_to_switch_in) modifies $filename,
# replacing any lines matching $pattern with $line_to_switch_in.
#
# It returns the number of lines it replaced (or would have replaced, if
# LOGONLY mode wasn't on...)
#
# B_replace_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
# Here an example of where you might use this:
#
# You'd like to replace any Options lines in Apache's config file with:
#            Options Indexes FollowSymLinks
#
###########################################################################

sub B_replace_line {
   
   my ($filename,$pattern,$line_to_switch_in) = @_;
   my $return=0;

   if ( &B_open_plus (*REPLACE_NEW,*REPLACE_OLD,$filename) ) {
      while (my $line=<REPLACE_OLD>) {
         unless ($line =~ $pattern) {    
            &B_print(*REPLACE_NEW,$line);
         } else {
            # Don't replace the line if it's already there.
            unless ($line eq $line_to_switch_in) {
               &B_print(*REPLACE_NEW,$line_to_switch_in);
               $return++;
               &ActionLog("# File modification in $filename -- replaced line\n");
               &ActionLog($line);
               &ActionLog("with:\n");
               &ActionLog("$line_to_switch_in");
            } else {
               # But if it is there, make sure it stays there! (by Paul Allen)
               &B_print(*REPLACE_NEW,$line);
            };    
         };
      }:
   &B_close_plus (*REPLACE_NEW,*REPLACE_OLD,$filename);
   } else {
      $return=0;
      &ErrorLog("Couldn't replace line(s) in $filename because open failed.\n");
   };

   $return;
};


###########################################################################
# &B_hash_comment_line ($filename,$pattern) modifies $filename, replacing 
# any lines matching $pattern with a "hash-commented" version, like this:
#
#
#        finger  stream  tcp     nowait  nobody  /usr/sbin/tcpd  in.fingerd
# becomes:
#        #finger  stream  tcp     nowait  nobody  /usr/sbin/tcpd  in.fingerd
#
#
# B_hash_comment_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
###########################################################################

sub B_hash_comment_line {

   my ($filename,$pattern) = @_;
   my $return=1;

   if ( &B_open_plus (*HASH_NEW,*HASH_OLD,$filename) ) {
      while (my $line=<HASH_OLD>) {
         unless ( ($line =~ $pattern) and ($line !~ /^\s*\#/) ) {    
            &B_print(*HASH_NEW,$line);
         } else {
            &B_print(*HASH_NEW,"#$line");
            &ActionLog("# File modification in $filename -- hash commented line\n");
            &ActionLog($line);
            &ActionLog("like this:\n");
            &ActionLog("#$line");
         };
      };
      &B_close_plus (*HASH_NEW,*HASH_OLD,$filename);
   } else {
      $return=0;
      &ErrorLog("Couldn't hash-comment line(s) in $filename because open failed.");
   };

   $return;
};


###########################################################################
# &B_hash_uncomment_line ($filename,$pattern) modifies $filename, 
# removing any commentting from lines that match $pattern.
#
#        #finger  stream  tcp     nowait  nobody  /usr/sbin/tcpd  in.fingerd
# becomes:
#        finger  stream  tcp     nowait  nobody  /usr/sbin/tcpd  in.fingerd
#
#
# B_hash_uncomment_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
###########################################################################

sub B_hash_uncomment_line {
   
   my ($filename,$pattern) = @_;
   my $return=1;

   if ( &B_open_plus (*HASH_NEW,*HASH_OLD,$filename) ) {
      while (my $line=<HASH_OLD>) {
         unless ( ($line =~ $pattern) and ($line =~ /^\s*\#/) ) {    
            &B_print(*HASH_NEW,$line);
         } else {
            $line =~ /^\s*\#+(.*)$/;
            $line = "$1\n";

            &B_print(*HASH_NEW,"$line");
            &ActionLog("File modification in $filename -- hash uncommented line\n");
            &ActionLog($line);
         };
      };
      &B_close_plus (*HASH_NEW,*HASH_OLD,$filename);
   } else {
      $return=0;
      &ErrorLog("Couldn't hash-uncomment line(s) in $filename because open failed.");
   };

   $return;
};



###########################################################################
# &B_delete_line ($filename,$pattern) modifies $filename, deleting any 
# lines matching $pattern.  It uses B_replace_line to do this.
#
# B_replace_line uses B_open_plus and B_close_plus, so that the file
# modified is backed up...
#
# Here an example of where you might use this:
#
# You'd like to remove any timeout=  lines in /etc/lilo.conf, so that your
# delay=1 modification will work.

#
###########################################################################


sub B_delete_line {

   my ($filename,$pattern)=@_;
   my $return=&B_replace_line($filename,$pattern,"");

   $return;
};


###########################################################################
# &B_chunk_replace ($file,$pattern,$replacement) reads $file replacing the
# first occurence of $pattern with $replacement.
# 
###########################################################################

sub B_chunk_replace {

   my ($file,$pattern,$replacement) = @_;

   my @lines;
   my $big_long_line;
   my $return=1;

   &B_open (*OLDFILE,$file);

   # Read all lines into one scalar.
   @lines = <OLDFILE>;
   &B_close (*OLDFILE);
   foreach my $line ( @lines ) {
      $big_long_line .= $line;
   };

   # Substitution routines get weird unless last line is terminated with \n
   chomp $big_long_line;
   $big_long_line .= "\n";

   # Exit if we don't find a match
   unless ($big_long_line =~ $pattern) {
      return 0;
   };
    
   $big_long_line =~ s/$pattern/$replacement/s;

   $return=&B_open_plus (*NEWFILE,*OLDFILE,$file);
   if ($return) {
      &B_print (*NEWFILE,$big_long_line);
      &B_close_plus (*NEWFILE,*OLDFILE,$file);
   };

   $return;
};


###########################################################################
# &B_chunk_append ($file,$pattern,$payload) reads $file adding the $payload
# after the first occurence of $pattern.
# 
###########################################################################

sub B_chunk_append {

   my ($file,$pattern,$payload) = @_;

   my @lines;
   my $big_long_line;
   my $return=1;
   my $position;

   my ($first_half,$second_half);

   &B_open(*OLDFILE,$file);

   # Read all lines into one scalar.
   @lines = <OLDFILE>;
   &B_close(*OLDFILE);
   foreach my $line ( @lines ) {
      $big_long_line .= $line;
   };

   # Substitution routines get weird unless last line is terminated with \n
   chomp $big_long_line;
   $big_long_line .= "\n";

   # Exit if we don't find a match
   unless ($big_long_line =~ $pattern) {
      return 0;
   };

   # We need to figure out which position in the file to insert the text
    
   $position = pos $big_long_line;

   # Insert the payload, unless it's already there.

   $first_half = substr ($big_long_line, 0, ($position-1) );
   $second_half = substr ($big_long_line, $position);

   my $length_to_check = length ($payload);
   unless ($legnth_to_check < 10) {
      $legnth_to_check =10;
   };
   unless ( substr($second_half,0,length ($payload)) eq $payload ) { 
      $big_long_line = $first_half . $payload . $second_half;
   };

   # Write out the result

   $return=&B_open_plus (*NEWFILE,*OLDFILE,$file);
   if ($return) {
      &B_print (*NEWFILE,$big_long_line);
      &B_close_plus (*NEWFILE,*OLDFILE,$file);
   };

   $return;
};



###########################################################################
# &B_delete_file ($file)  deletes the file $file and makes a backup to
# the backup directory.
#
# Here an example of where you might use this:
#
# You'd like to add a   Options Indexes  line to Apache's config. file,
# after you delete all Options lines from said config file.
#
##########################################################################


sub B_delete_file {

   #
   # This API routine deletes the named file, backing it up first to the
   # backup directory.
   # 

   my $filename=shift @_;
   my $return=1;

   # We have to append the prefix ourselves since we don't use B_open_plus

   my $original_filename=$filename;
   $filename .= $GLOBAL_PREFIX;

   &ActionLog("# Deleting (and backing-up) file $original_filename\n");
   &ActionLog("rm $original_filename\n");

   unless ($filename) {
      &ErrorLog("# B_delete_file called with no arguments!\n");
   };
    
   unless ($GLOBAL_LOGONLY) {
      if ( B_backup_file($original_filename) ) {
         unless ( unlink $filename ) {
            &ErrorLog("# Couldn't unlink file $original_filename");
            $return=0;
         };
      } else {
         $return=0;
         &ErrorLog("# B_delete_file did not delete $original_filename since it could not back it up\n");
      }:
   };

   $return;

};


###########################################################################
# &B_create_file ($file) creates the file $file, if it doesn't already
# exist.
# It will set a default mode of 0700 and a default uid/gid or 0/0.
#
# &B_create_file, to support Bastille's undo functionality, writes an
# rm $file command to the end of the file $GLOBAL_BFILE{"created-files"}.
#
##########################################################################


sub B_create_file {

   my $file = $_[0];
   my $return=1;

   # We have to create the file ourselves since we don't use B_open_plus

   my $original_file = $file;
   $file .= $GLOBAL_PREFIX;

   unless ( -e $file ) {
      unless ($GLOBAL_LOGONLY) {
         $return=open CREATE_FILE,">$file";
       
         if ($return) {
            close CREATE_FILE;
            chmod 0700,$file;

            # Make the undo functionality
            &B_init_file ("created-files");
      
            open CREATED_FILE,">>" . $GLOBAL_BFILE{"created-files"};
            print CREATED_FILE "rm $original_file\n";
            close CREATED_FILE;

         } else {
            &ErrorLog("# Couldn't create file $original_file even though it didn't already exist!");
         };
       
      };

      &ActionLog("# Created file $original_file\n");

   } else {
      &ActionLog("# Didn't create file $original_file since it already existed.\n");
      $return = 0;
   };

   $return;
};

       
###########################################################################
# &B_create_dir ($dir) creates the directory $dir, if it doesn't already
# exist.
# It will set a default mode of 0700 and a default uid/gid or 0/0.
#
# &B_create_dir, to support Bastille's undo functionality, writes an
# rmdir $file command to the end of the file $GLOBAL_BFILE{"created-files"}.
#
##########################################################################


sub B_create_dir {

   my $dir = $_[0];
   my $return = 1;

   # We have to append the prefix ourselves since we don't use B_open_plus

   my $original_dir = $dir;
   $dir .= $GLOBAL_PREFIX;

   unless ( -d $dir ) {
      unless ($GLOBAL_LOGONLY) {
         $return = mkdir ($dir,0700);
       
         if ($return) {

            # Make the undo functionality
            &B_init_file ("created-dirs");
      
            open CREATED_DIR,">>" . $GLOBAL_BFILE{"created-dirs"};
            print CREATED_DIR "rmdir $original_dir\n";
            close CREATED_DIR;

         } else {
            &ErrorLog("# Couldn't create dir $original_dir even though it didn't already exist!");
         };
       
      };
      &ActionLog("# Created directory $original_dir\n");
   } else {
      &ActionLog("# Didn't create directory $original_dir since it already existed.\n");
      $return=0;
   };

   $return;
};
      
       
###########################################################################
# &B_print ($handle,@list) prints the items of @list to the filehandle
# $handle.  It logs the action and respects the $GLOBAL_LOGONLY variable.
#
###########################################################################

sub B_print {
   my $handle=shift @_;

   my $result=1;

   unless ($GLOBAL_LOGONLY) {
      $result = print $handle @_;
   };

   ($handle) = "$handle" =~ /[^:]+::[^:]+::([^:]+)/;

   $result;
};


###########################################################################
# &B_chmod ($mode, $file) sets the mode of $file to $mode.  $mode must
# be stored in octal, so if you want to give mode 700 to /etc/aliases,
# you need to use:
#
#                 &B_chmod ( 0700 , "/etc/aliases");
#
# where the 0700 denotes "octal 7-0-0".
#
# &B_chmod respects GLOBAL_LOGONLY and the $GLOBAL_PREFIX and uses 
# &B_backup to save an original copy of the file (with the right 
# permissions.)
#
##########################################################################


sub B_chmod {
   my ($new_perm,$file)=@_;
   my $old_perm;
   my $old_perm_raw;
   my $new_perm_formatted;

   my $return=1;

   # Prepend global prefix, but save a the original filename for B_backup_file
   my $original_file=$file;
   $file .= $GLOBAL_PREFIX;

   # Store the old permissions so that we can log them.
   $old_perm_raw=(stat $file)[2];   
   $old_perm = ( ($old_perm_raw/64) % 8) .
      ( ($old_perm_raw/8) % 8) . 
      ($old_perm_raw % 8);

   $new_perm_formatted = sprintf "%5lo",$new_perm;
   
   &ActionLog ("# change permissions on $original_file from $old_perm to $new_perm_formatted\n");

   &ActionLog( "chmod $new_perm_formatted,\"$original_file\";\n");

   # Change the permissions on the file

   if ( -e $file ) {
      unless ($GLOBAL_LOGONLY) {
         &B_backup_file ($original_file);
         $return=chmod $new_perm,$file;
      };
      unless ($return) {
         &ActionLog("#ERROR: couldn't change permissions on $original_file from $old_perm to $new_perm_formatted\n");
         &ErrorLog("#ERROR: couldn't change permissions on $original_file from $old_perm to $new_perm_formatted\n");
        
         $return=0;
      };
   } else {
      &ActionLog( "#ERROR: chmod: File $original_file doesn't exist!\n");
      &ErrorLog( "#ERROR: chmod: File $original_file doesn't exist!\n");    
      $return=0;
   };
  
   $return;

};


###########################################################################
# &B_chown ($uid, $file) sets the owner of $file to $uid, like this:
#
#                 &B_chown ( 0 , "/etc/aliases");
#
# &B_chown respects $GLOBAL_LOGONLY and the $GLOBAL_PREFIX and uses 
# &B_backup to save an original copy of the file (with the right 
# owner.)
#
# Unlike Perl, we've broken the chown function into B_chown/B_chgrp to
# make error checking simpler.
#
##########################################################################


sub B_chown {
   my ($newown,$file)=@_;
   my $oldown;
   my $oldgown;

   my $return=1;
   
   # Prepend prefix, but save original filename for B_backup_dir
   my $original_file=$file;
   $file .= $GLOBAL_PREFIX;

   $oldown=(stat $file)[4];
   $oldgown=(stat $file)[5];

   &ActionLog ("# change ownership on $original_file from $oldown to $newown\n");
   &ActionLog ("chown $newown,$oldgown,\"$original_file\";\n");
   if ( -e $file ) {
      unless ($GLOBAL_LOGONLY) {
         &B_backup_file ($original_file);
         $return = chown $newown,$oldgown,$file;
      };
      unless ($return) {
         &ActionLog( "#ERROR: couldn't change ownership to $newown on file $original_file\n");
         &ErrorLog("#ERROR: couldn't change ownership to $newown on file $original_file\n");
      };
   } else {
      &ActionLog("#ERROR: chown: File $original_file doesn't exist!\n");
      &ErrorLog("#ERROR: chown: File $original_file doesn't exist!\n");
      $return = 0;
   };

   $return;
};




###########################################################################
# &B_chgrp ($gid, $file) sets the group owner of $file to $gid, like this:
#
#                 &B_chgrp ( 0 , "/etc/aliases");
#
# &B_chgrp respects $GLOBAL_LOGONLY and the $GLOBAL_PREFIX and uses 
# &B_backup to save an original copy of the file (with the right 
# owner.)
#
# Unlike Perl, we've broken the chown function into B_chown/B_chgrp to
# make error checking simpler.
#
##########################################################################


sub B_chgrp {
   my ($newgown,$file)=@_;
   my $oldown;
   my $oldgown;

   my $return=1;
   
   # Prepend global prefix, but save original filename for &B_backup_file
   my $original_file=$file;
   $file .= $GLOBAL_PREFIX;

   $oldown=(stat $file)[4];
   $oldgown=(stat $file)[5];

   &ActionLog( "# change group ownership on $original_file from $oldgown to $newgown\n");
   &ActionLog( "chown $oldown,$newgown,\"$original_file\";\n");
   if ( -e $file ) {
      unless ($GLOBAL_LOGONLY) {
         &B_backup_file($original_file);
         $return = chown $oldown,$newgown,$file;
      };
      unless ($return) {
         &ActionLog("#ERROR: couldn't change ownership to $newgown on file $original_file\n");
         &ErrorLog("#ERROR: couldn't change ownership to $newgown on file $original_file\n");       
      };
   } else {
      &ActionLog( "#ERROR: chgrp: File $original_file doesn't exist!\n");
      &ErrorLog("#ERROR: chgrp: File $original_file doesn't exist!\n");
      $return=0;
   };

   $return;
};


###########################################################################
# &B_symlink ($original_file,$new_symlink) creates a symbolic link from
# $original_file to $new_symlink.
#
# &B_symlink respects $GLOBAL_LOGONLY and the $GLOBAL_PREFIX.  It supports
# the undo functionality that you've come to know and love by adding every
# symbolic link it creates to $GLOBAL_BFILE{"created-symlinks"}, currently set to:
#
#         /root/Bastille/undo/undo-created-symlinks
#
# The undo script, if it works like I think it should, will run this file,
# which should be a script or rm's...
#
##########################################################################

sub B_symlink {
   my ($source_file,$new_symlink)=@_;

   my $return=1;

   my $original_source = $source_file;
   my $original_symlink = $new_symlink;
    
   $source_file .= $GLOBAL_PREFIX;
   $new_symlink .= $GLOBAL_PREFIX;

   unless ($GLOBAL_LOGONLY) {
      $return=symlink $source_file,$new_symlink;
      if ($return) {
         &B_init_file ("created-symlinks");

         open SYMLINK_FILE,">>" . $GLOBAL_BFILE{"created-symlinks"};
         print SYMLINK_FILE "rm $original_symlink\n";
         close SYMLINK_FILE;
      };
   };

   &ActionLog( "# created a symbolic link called $original_symlink from $original_source\n");
   &ActionLog( "symlink \"$original_source\",\"$original_symlink\";\n");
   unless ($return) { 
      &ActionLog("#ERROR: couldn't symlink $original_symlink -> $original_source\n");
      &ErrorLog("#ERROR: couldn't symlink $original_symlink -> $original_source\n");
   };

   $return;
};


###########################################################################
# B_Display was the display function used by IPCHAINS in v1.0.0-v1.0.3.
# Its use should be fully deprecated in 1.1.
#
###########################################################################

sub B_Display {
   # routine to display a long string, automatically wrapping
   # each line to no more than 80 characters
   my ( $text) = @_;
   my ($chunk1, $chunk2, $chunk3);
   while ( length($text) >= 80 ) {
      # pull the first 80 chars off the string
      $text =~ s/^(.{80})//s;
      $chunk1 = $1;
      if ( $chunk1 =~ /^(.*?\n)(.*?)$/s ) {
         $chunk2 = $1; $chunk3 = $2;
         # explicit line endings should be respected
         print $chunk2;
         $text = $chunk3 . $text;
      } elsif ( $chunk1 =~ /\s/ ) {
         # there's a space, grab everything up to the _last_ space
         $chunk1 =~ /^(.*)\s([^\s]*?)$/s;
         $chunk2 = $1; $chunk3 = $2;
         # print the stuff up to the last space
         print "$chunk2\n";
         # put the remaining data back at the front of $text
         $text = $chunk3 . $text;
      } else {
         # no space, just print
         print "$chunk1\n";
      };
   };
   if ( $text =~ /\S/ ) {
      # some non-whitespace data, print it
      print "$text\n";
   };
};


###########################################################################
# &B_chkconfig_off ($deamon_name) deletes the symbolic links that are
# named in the "# chkconfig: ___ _ _ " portion of the init.d files.  We
# need this utility, in place of the distro's chkconfig, because of both
# our need to add undo functionality and our need to harden distros that
# are not mounted on /.
#
# chkconfig allows for an UNDO of its work by writing to an executable
# file $GLOBAL_BFILE{"removed-symlinks"}.
#
# It uses the following global variables to find the links and the init
# scripts, respectively:
#
#   $GLOBAL_DIR{"rcd"}    -- directory where the rc_.d subdirs can be found
#   $GLOBAL_DIR{"initd"}  -- directory the rc_.d directories link to
#
# Here an example of where you might use this:
#
# You'd like to tell stop running sendmail in daemon mode on boot:
#       B_chkconfig_off("sendmail")
#
###########################################################################



sub B_chkconfig_off {

   my $startup_script=$_[0];
   my $return=1;

   my $chkconfig_line;
   my @runlevels;
   my ($start_order,$stop_order);

   # Run through the init script looking for the chkconfig line...

   $return = open CHKCONFIG,$GLOBAL_DIR{"initd"} . "/$startup_script";
   unless ($return) {
      &ActionLog("Didn't chkconfig_off $startup_script because we couldn't open " . $GLOBAL_DIR{"initd"} . "/$startup_script\n");
   } else {

      READ_LOOP:
      while (my $line=<CHKCONFIG>) {

         # We're looking for lines like this one:
         #      # chkconfig: 2345 10 90
       
         if ($line =~ /^#\s*chkconfig:\s*(\d+)\s*(\d+)\s*(\d+)/ ) {
            @runlevels=split //,$1;
            $start_order=$2;
            $stop_order=$3;
            last READ_LOOP;
         };
      };
      close CHKCONFIG;
   
      # If we never found a chkconfig line, can we just run through all 5 
      # rcX.d dirs from 1 to 5...?
   
      # unless ( $start_order and $stop_order ) {
      #    @runlevels=("1","2","3","4","5");
      #    $start_order = "*"; $stop_order="*";
      # };
   
      # Now, run through removing symlinks...
   
      $return=0;

      foreach my $level ( @runlevels ) {
         my $link = $GLOBAL_DIR{"rcd"} . "/rc" . $level . ".d/S$start_order" . $startup_script;
         my $target = $GLOBAL_DIR{"initd"} ."/" . $startup_script;
         my $local_return;
       
         if ( -e $link ) {
            unless ($GLOBAL_LOGONLY) {
               $local_return=unlink $link;
            };
            if ($local_return) {
               $return++;
               &ActionLog("# Removed link $link\n");
          
               #
               # If we removed the link, add a link command to the undo file
               #
          
               &B_init_file ("removed-symlinks");

               open REMOVED_SYMLINKS,">>" . $GLOBAL_BFILE{"removed-symlinks"};
               print REMOVED_SYMLINKS "ln -s $target $link\n";
               close REMOVED_SYMLINKS;
          
            } else {
               &ErrorLog("Couldn't remove $link when trying to chkconfig off $startup_script\n");
            };
      
         };
      };
   
   };

   if ($return < @runlevels) {
      $return=0;
   };

   $return;

};


############################################################################
# &B_cp is the Bastille cp command, which is based on Perl's File::cp.
# &B_cp($source,$target).  It is somewhat strange, to make the Backup and
# undo functions easier to implement, in that:
#
#
#     It can ONLY copy from one file to another! Both $source and
#     $target must be files, not directories!!!!
#
# It respects $GLOBAL_LOGONLY and $GLOBAL_PREFIX, appending the prefix for
# itself.  If $target is an already-existing file, it is backed up.
#
# Undo either appends another  rm $target   to  $GLOBAL_BFILE{"created-files"}  # or backs up the file that _was_ there into the $GLOBAL_BDIR{"backup"}.
#
############################################################################

sub B_cp {

   my ($source,$target)=@_;
   my $return=0;

   my $had_to_backup_target=0;

   use File::Copy;

   my $original_source = $source;
   $source = $GLOBAL_PREFIX . $source;
   my $original_target = $target;
   $target = $GLOBAL_PREFIX . $target;

   unless ($GLOBAL_LOGONLY) {
      unless ( -e $target and -f $target ) {
         &B_backup_file($original_target);
         &ErrorLog("About to copy $original_source to $original_target -- had to backup target\n");
         $had_to_backup_target=1;
      };

      $return=copy($source,$target);
      if ($return) {
         &ActionLog("cp $original_source $original_target\n");
       
         #
         # We want to add a line to the $GLOBAL_BFILE{"created-files"} so that the
         # file we just put at $original_target gets deleted.
         #
       
         &B_init_file ("created-files");

         open CREATED,">>" . $GLOBAL_BFILE{"created-files"};
         print CREATED "rm $original_target\n";
         close CREATED;
       
      } else {
         &ErrorLog("Failed to copy $original_source to $original_target\n");
      };
   };

   $return;

};



############################################################################
# &B_place puts a file in place, using Perl's File::cp.  This file is taken
# from $GLOBAL_BDIR{"install"} and is used to place a file that came with
# Bastille.  This is a workaround to a problem with using &B_cp for the job:
# &B_cp prepends a GLOBAL_PREFIX to both filenames, while &B_place only 
# prepends to the destination filename.
#
#
# It respects $GLOBAL_LOGONLY and $GLOBAL_PREFIX, appending the prefix for
# itself.  If $target is an already-existing file, it is backed up.
#
# Undo either appends another  rm $target   to  $GLOBAL_BFILE{"created-files"}  or
# backs up the file that _was_ there into the $GLOBAL_BDIR{"backup"}.
#
############################################################################

sub B_place {

   my ($source,$target)=@_;
   my $return=0;

   my $had_to_backup_target=0;

   use File::Copy;

   my $original_source=$source;
   $source  = $GLOBAL_BDIR{"install"} . $source;
   my $original_target=$target;
   $target = $GLOBAL_PREFIX . $target;

   unless ($GLOBAL_LOGONLY) {
      unless ( -e $target and -f $target ) {
         &B_backup_file($original_target);
         &ErrorLog("About to copy $original_source to $original_target -- had to backup target\n");
         $had_to_backup_target=1;
      };

      $return = copy($source,$target);
      if ($return) {
         &ActionLog("placed file $original_source  as  $original_target\n");
       
         #
         # We want to add a line to the $GLOBAL_BFILE{"created-files"} so that the
         # file we just put at $original_target gets deleted.
         #
       
         &B_init_file ("created-files");

         open CREATED,">>" . $GLOBAL_BFILE{"created-files"};
         print CREATED "rm $original_target\n";
         close CREATED;
      } else {
         &ErrorLog("Failed to place $original_source as $original_target\n");
      };
   };

   $return;
};


#############################################################################
#############################################################################
#############################################################################

###########################################################################
# &B_mknod ($file) creates the node $file, if it doesn't already
# exist.  It uses the prefix and suffix, like this:
#
#            mknod $prefix $file $suffix
#
# This is just a wrapper to the mknod program, which tries to introduce
# undo functionality, by writing    rm $file     to the end of the 
# file $GLOBAL_BFILE{"created-files"}.
#
##########################################################################


sub B_mknod {

   my ($prefix,$file,$suffix) = @_;
   my $return=1;

   # We have to create the filename ourselves since we don't use B_open_plus

   my $original_file = $file;
   $file .= $GLOBAL_PREFIX;

   unless ( -e $file ) {
      unless ($GLOBAL_LOGONLY) {
         $return=`$GLOBAL_BIN{"mknod"} $prefix $file $suffix`;
       
         if ($return) {

            # Make the undo functionality
            &B_init_file ("created-files");
      
            open CREATED_FILE,">>" . $GLOBAL_BFILE{"created-files"};
            print CREATED_FILE "rm $original_file\n";
            close CREATED_FILE;

         } else {
            &ErrorLog("# Couldn't mknod $prefix $original_file $suffix even though it didn't already exist!");
         };
       
      };
      &ActionLog("# mknod $prefix $original_file $suffix\n");
   } else {
      &ActionLog("# Didn't mknod $prefix $original_file $suffix since $original_file already existed.\n");
      $return=0;
   };

   $return;
};

1;
